/*	Abreast Davies-Meyer Implementation employing Serpent
	Copyright (C) 1999, Daniel Roethlisberger

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	The author of this program may be contacted at admin@roe.ch.
*/

#include <windows.h>
#include <memory.h>
#include <malloc.h>
#include "hash.h"
#include "types.h"
#include "serpent.h"

void HashInit(HASH_CTX* ctx)
{
	ctx->hash.A = 0x31337CDC;
	ctx->hash.B = 0x31337CDC;
	ctx->hash.C = 0x31337CDC;
	ctx->hash.D = 0x31337CDC;
	ctx->hash.E = 0x31337CDC;
	ctx->hash.F = 0x31337CDC;
	ctx->hash.G = 0x31337CDC;
	ctx->hash.H = 0x31337CDC;
	ctx->len = 0;
	ctx->data = (unsigned char*)malloc(16);
	return;
}

void HashUpdate(HASH_CTX* ctx, unsigned char* buffer, unsigned int len)
{
	int blocks;
	int chars;

	chars = len + ctx->len;
	blocks = chars >> 4;
	
	if(blocks)
	{
		if(ctx->len)
		{
			memcpy(ctx->data + ctx->len, buffer, 16 - ctx->len);
			HashTransform(ctx, ctx->data);
			blocks--;
		}
		else
			ctx->len = 16;
		for(int i = 0; i < blocks; i++)
		{
			HashTransform(ctx, buffer + 16 - ctx->len + i*16);
		}
		ctx->len = chars % 16;
		if(ctx->len)
			memcpy(ctx->data, buffer + len - ctx->len, ctx->len);
	}
	else
	{
		memcpy(ctx->data + ctx->len, buffer, len);
		ctx->len += len;
	}

	return;
}

void HashFinal(unsigned char digest[32], HASH_CTX* ctx)
{
	if(ctx->len)
	{
		memset(ctx->data + ctx->len, 0x00, 16 - ctx->len);
		HashTransform(ctx, ctx->data);
	}
	memcpy(digest, (BYTE*)&ctx->hash, 32);
	free(ctx->data);
	return;
}

void HashTransform(HASH_CTX* ctx, unsigned char block[16])
{
	BETA _H;				// copy of H[i-1]
	BETA _G;				// copy of G[i-1]

	u4byte keyctx[140];		// Serpent internal key schedule
	u4byte key[8];			// Serpent 256 bit user key

	// save old G and H
	memcpy((BYTE*)&_H, (BYTE*)&ctx->H, 16);
	memcpy((BYTE*)&_G, (BYTE*)&ctx->G, 16);
	
	// process H
	memcpy((BYTE*)key, (BYTE*)&_G, 16);
	memcpy(((BYTE*)key) + 16, block, 16);
	SerpentKeyInit(key, keyctx);
	SerpentEncrypt((u4byte*)&ctx->H, keyctx);
	ctx->H.A ^= _H.A;
	ctx->H.B ^= _H.B;
	ctx->H.C ^= _H.C;
	ctx->H.D ^= _H.D;

	// process G
	memcpy((BYTE*)key, block, 16);
	memcpy(((BYTE*)key) + 16, (BYTE*)&_H, 16);
	SerpentKeyInit(key, keyctx);
	SerpentEncrypt((u4byte*)&ctx->G, keyctx);
	ctx->G.A ^= _G.A;
	ctx->G.B ^= _G.B;
	ctx->G.C ^= _G.C;
	ctx->G.D ^= _G.D;

	return;
}

